from typing import Optional

import click

from mlfoundry import env_vars
from mlfoundry.login import login as login_


@click.command(short_help="Store API key in the local file system")
@click.option(
    "--host",
    "--tracking-uri",
    "--tracking_uri",
    "tracking_uri",
    type=click.STRING,
    required=True,
    envvar=env_vars.TRACKING_HOST_GLOBAL,
    help="Tracking server host (URL to Truefoundry platform)",
)
@click.option(
    "--relogin",
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite existing API key for the given `--host`/`--tracking_uri`",
)
@click.option(
    "--api-key",
    "--api_key",
    type=click.STRING,
    default=None,
    prompt="Please enter an API Key",
    hide_input=True,
    prompt_required=False,
)
def login(tracking_uri: Optional[str], relogin: bool, api_key: Optional[str]):
    login_(tracking_uri=tracking_uri, relogin=relogin, api_key=api_key)



import copy
import datetime
import json
import logging
import os.path
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from mlflow.entities import ArtifactType, CustomMetric, Metric, Model, ModelSchema
from mlflow.entities import ModelVersion as _ModelVersion
from mlflow.tracking import MlflowClient
from mlflow.transformers import _get_or_infer_task_type

from mlfoundry.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from mlfoundry.enums import ModelFramework
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import (
    FILES_DIR,
    INTERNAL_METADATA_PATH,
    MODEL_DIR_NAME,
    MODEL_SCHEMA_UPDATE_FAILURE_HELP,
)
from mlfoundry.log_types.artifacts.utils import (
    _copy_additional_files,
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
)
from mlfoundry.pydantic_v1 import BaseModel, Extra
from mlfoundry.version import __version__

logger = logging.getLogger("mlfoundry")


# TODO: Add some progress indicators for upload and download
# TODO: Support async download and upload


class ModelVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root
    model_dir: str  # relative to `files_dir`
    model_is_null: bool = False
    framework: ModelFramework = ModelFramework.UNKNOWN
    transformers_pipeline_task: Optional[str] = None
    model_filename: Optional[str] = None
    mlfoundry_version: Optional[str] = None

    def dict(self, *args, **kwargs):
        dct = super().dict(*args, **kwargs)
        dct["framework"] = dct["framework"].value
        return dct


class ModelVersionDownloadInfo(BaseModel):
    download_dir: str
    model_dir: str
    model_framework: ModelFramework = ModelFramework.UNKNOWN
    model_filename: Optional[str] = None


class ModelVersion:
    def __init__(
        self,
        model_version: _ModelVersion,
        model: Model,
    ) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._model_version: _ModelVersion = model_version
        self._model: Model = model
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._model_schema: Optional[ModelSchema] = None
        # TODO (chiragjn): The default on `model_version.metrics` is `list` there is no way to
        #   distinguish if the API returned something or there are no metrics
        self._metrics = None
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str):
        """
        Get the version of a model to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Returns:
            ModelVersion: An ModelVersion instance of the Model

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = mlfoundry.ModelVersion.from_fqn(fqn="<your-model-fqn>")
            ```
        """
        mlflow_client = _get_mlflow_client()
        model_version = mlflow_client.get_model_version_by_fqn(fqn=fqn)
        model = mlflow_client.get_model_by_id(model_id=model_version.model_id)
        instance = cls(model_version=model_version, model=model)
        instance._metrics = model_version.metrics or []
        return instance

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Model Version was deleted, cannot perform updates on a deleted version"
            )

    def _refetch_model_version(self):
        self._model_version = self._mlflow_client.get_model_version_by_id(
            version_id=self._model_version.id
        )

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._refetch_model_version()
        self._description = self._model_version.description or ""
        self._metadata = copy.deepcopy(self._model_version.artifact_metadata)
        self._model_schema = copy.deepcopy(self._model_version.model_schema)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r}, model_schema={self._model_schema!r})"

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=self._model_version.id
            ),
            mlflow_client=self._mlflow_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the model"""
        return self._model.name

    @property
    def model_fqn(self) -> str:
        """Get fqn of the model"""
        return self._model.fqn

    @property
    def version(self) -> int:
        """Get version information of the model"""
        return self._model_version.version

    @property
    def fqn(self) -> str:
        """Get fqn of the current model version"""
        return self._model_version.fqn

    @property
    def step(self) -> int:
        """Get the step in which model was created"""
        return self._model_version.step

    @property
    def description(self) -> Optional[str]:
        """Get description of the model"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the model"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current model"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current model"""
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def model_schema(self) -> Optional[ModelSchema]:
        """get the model schema for current model"""
        return self._model_schema

    @model_schema.setter
    def model_schema(self, value: Union[Dict[str, Any], ModelSchema]):
        if not isinstance(value, ModelSchema):
            value = ModelSchema.parse_obj(value)
        self._model_schema = value

    @property
    def metrics(self) -> Dict[str, Union[float, int]]:
        """get the metrics for the current version of the model"""
        if self._metrics is None:
            self._refetch_model_version()
            metrics_as_kv = {}
            metrics: List[Metric] = sorted(
                self._model_version.metrics or [], key=lambda m: m.timestamp
            )
            for metric in metrics:
                metrics_as_kv[metric.key] = metric.value
            self._metrics = metrics_as_kv
        return self._metrics

    @property
    def created_by(self) -> str:
        """Get the information about who created the model version"""
        return self._model_version.created_by

    @property
    def created_at(self) -> datetime.datetime:
        """Get the time at which model version was created"""
        return self._model_version.created_at

    @property
    def updated_at(self) -> datetime.datetime:
        """Get the information about when the model version was updated"""
        return self._model_version.updated_at

    def raw_download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an model file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified models. This directory must already exist.
                        If unspecified, the models will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired models.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.raw_download(path="<your-desired-download-path>")
            ```
        """
        logger.info("Downloading model version contents, this might take a while ...")
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.download_artifacts(
            artifact_path="", dst_path=path, overwrite=overwrite, progress=progress
        )

    def _download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> Tuple[ModelVersionInternalMetadata, ModelVersionDownloadInfo]:
        self._ensure_not_deleted()
        download_dir = self.raw_download(
            path=path, overwrite=overwrite, progress=progress
        )
        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                f"Model version seems to be corrupted or in invalid format due to missing model metadata. "
                f"You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ModelVersionInternalMetadata.parse_obj(json.load(f))
        download_info = ModelVersionDownloadInfo(
            download_dir=os.path.join(download_dir, internal_metadata.files_dir),
            model_dir=os.path.join(
                download_dir, internal_metadata.files_dir, internal_metadata.model_dir
            ),
            model_framework=internal_metadata.framework,
            model_filename=internal_metadata.model_filename,
        )
        return internal_metadata, download_info

    def download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> ModelVersionDownloadInfo:
        """
        Download a model file or directory to a local directory if applicable, and return download info
        containing `model_dir` - local path where model was downloaded

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified models. This directory must already exist.
                        If unspecified, the models will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            ModelVersionDownloadInfo:  Download Info instance containing
                `model_dir` (path to downloaded model folder) and other metadata

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            download_info = model_version.download(path="<your-desired-download-path>")
            print(download_info.model_dir)
            ```
        """
        _, download_info = self._download(
            path=path, overwrite=overwrite, progress=progress
        )
        return download_info

    def load(self, **load_model_kwargs):
        from mlfoundry.frameworks import get_model_registry

        internal_metadata, download_info = self._download(path=None)
        if internal_metadata.model_is_null:
            raise MlFoundryException(
                f"Cannot load model as no model object was provided during logging."
            )
        if internal_metadata.framework == ModelFramework.UNKNOWN:
            raise MlFoundryException(
                "Cannot deserialize model as model framework is unknown. You can still download the "
                "contents by calling `.download(path='your/path/here') and load the model manually."
            )
        return get_model_registry(internal_metadata.framework).load_model(
            download_info.model_dir, **load_model_kwargs
        )

    def delete(self) -> bool:
        """
        Deletes the current instance of the ModelVersion hence deleting the current version.

        Returns:
            True if model was deleted successfully

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.delete()
            ```
        """
        self._ensure_not_deleted()
        self._mlflow_client.delete_artifact_version(version_id=self._model_version.id)
        self._deleted = True
        return True

    def update(self):
        """
        Updates the current instance of the ModelVersion hence updating the current version.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.description = 'This is the new description'
            model_version.update()
            ```
        """
        self._ensure_not_deleted()
        kwargs = {}
        if self.model_schema is not None:
            kwargs["model_schema"] = self.model_schema
        self._model_version = self._mlflow_client.update_model_version(
            version_id=self._model_version.id,
            description=self.description,
            artifact_metadata=self.metadata,
            **kwargs,
        )
        self._set_mutable_attrs()


def calculate_model_size(artifact_dir: tempfile.TemporaryDirectory):
    """
    Tells about the size of the model

    Args:
        artifact_dir (str): directory in which model is present.

    Returns:
        total size of the model
    """
    total_size = 0
    for path, dirs, files in os.walk(artifact_dir.name):
        for f in files:
            file_path = os.path.join(path, f)
            total_size += os.stat(file_path).st_size
    return total_size


def _log_model_version(
    run,
    name: str,
    model: Any,
    framework: ModelFramework,
    model_file_or_folder: Optional[str] = None,
    mlflow_client: Optional[MlflowClient] = None,
    ml_repo_id: Optional[str] = None,
    model_save_kwargs: Optional[Dict[str, Any]] = None,
    additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]] = (),
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    model_schema: Optional[Union[Dict[str, Any], ModelSchema]] = None,
    custom_metrics: Optional[Union[List[Dict[str, Any]], CustomMetric]] = None,
    step: Optional[int] = 0,
    progress: Optional[bool] = None,
) -> ModelVersion:
    from mlfoundry.frameworks import get_model_registry

    if (run and mlflow_client) or (not run and not mlflow_client):
        raise MlFoundryException("Exactly one of run, mlflow_client should be passed")
    if mlflow_client and not ml_repo_id:
        raise MlFoundryException(
            "If mlflow_client is passed, ml_repo_id must also be passed"
        )
    if run:
        mlflow_client: MlflowClient = run.mlflow_client

    custom_metrics = custom_metrics or []
    model_save_kwargs = model_save_kwargs or {}
    metadata = metadata or {}
    additional_files = additional_files or {}
    step = step or 0

    # validations
    if framework is None:
        framework = ModelFramework.UNKNOWN
    elif not isinstance(framework, ModelFramework):
        framework = ModelFramework(framework)

    if (
        model_file_or_folder is None
        and model is not None
        and framework == ModelFramework.UNKNOWN
    ):
        raise MlFoundryException(
            "`framework` cannot be passed as 'unknown' when `model` is not None"
        )

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    if model_schema is not None and not isinstance(model_schema, ModelSchema):
        model_schema = ModelSchema.parse_obj(model_schema)

    if custom_metrics and not model_schema:
        raise MlFoundryException(
            "Custom Metrics defined without adding the Model Schema"
        )
    custom_metrics = [
        CustomMetric.parse_obj(cm) if not isinstance(cm, CustomMetric) else cm
        for cm in custom_metrics
    ]

    logger.info("Logging model and additional files, this might take a while ...")
    temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

    internal_metadata = ModelVersionInternalMetadata(
        framework=framework,
        files_dir=FILES_DIR,
        model_dir=MODEL_DIR_NAME,
        model_is_null=model is None,
        model_filename=(
            os.path.basename(model_file_or_folder)
            if model_file_or_folder and os.path.isfile(model_file_or_folder)
            else None
        ),
        mlfoundry_version=__version__,
    )

    try:
        local_files_dir = os.path.join(temp_dir.name, internal_metadata.files_dir)
        os.makedirs(local_files_dir, exist_ok=True)
        # in case model was None, we still create an empty dir
        local_model_dir = os.path.join(local_files_dir, internal_metadata.model_dir)
        os.makedirs(local_model_dir, exist_ok=True)
        if not model_file_or_folder:
            # serialize model
            logger.info("Serializing model files to model version contents")
            local_model_dir = os.path.join(local_files_dir, internal_metadata.model_dir)
            if model is not None:
                model_registy = get_model_registry(framework)
                model_registy.save_model(
                    model=model, path=local_model_dir, **model_save_kwargs
                )
                if framework == ModelFramework.TRANSFORMERS:
                    internal_metadata.transformers_pipeline_task = (
                        _get_or_infer_task_type(model, model_save_kwargs.get("task"))
                    )
        else:
            logger.info("Adding model file/folder to model version content")
            model_file_or_folder = [
                (model_file_or_folder, MODEL_DIR_NAME.rstrip(os.sep) + os.sep)
            ]
            _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir=internal_metadata.files_dir,
                model_dir=internal_metadata.model_dir,
                additional_files=model_file_or_folder,
                ignore_model_dir_dest_conflict=True,
            )

        # verify additional files and paths, copy additional files
        if additional_files:
            logger.info("Adding `additional_files` to model version contents")
            _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir=internal_metadata.files_dir,
                model_dir=internal_metadata.model_dir,
                additional_files=additional_files,
                ignore_model_dir_dest_conflict=False,
            )

    except Exception as e:
        temp_dir.cleanup()
        raise MlFoundryException("Failed to log model") from e

    # save internal metadata
    local_internal_metadata_path = os.path.join(temp_dir.name, INTERNAL_METADATA_PATH)
    os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
    with open(local_internal_metadata_path, "w") as f:
        json.dump(internal_metadata.dict(), f)

    # create entry
    version_id = mlflow_client.create_artifact_version(
        experiment_id=int(run._experiment_id) if run else ml_repo_id,
        artifact_type=ArtifactType.MODEL,
        name=name,
    )
    artifacts_repo = MlFoundryArtifactsRepository(
        artifact_identifier=ArtifactIdentifier(artifact_version_id=version_id),
        mlflow_client=mlflow_client,
    )
    model_size = calculate_model_size(temp_dir)
    try:
        logger.info(
            "Packaging and uploading files to remote with Total Size: %.6f MB",
            model_size / 1000000.0,
        )
        artifacts_repo.log_artifacts(
            local_dir=temp_dir.name, artifact_path=None, progress=progress
        )
    except Exception as e:
        mlflow_client.notify_failure_for_artifact_version(version_id=version_id)
        raise MlFoundryException("Failed to log model") from e
    finally:
        temp_dir.cleanup()
    mlflow_client.finalize_artifact_version(
        version_id=version_id,
        run_uuid=run.run_id if run else None,
        artifact_size=model_size,
        step=step if run else None,
    )
    model_version = mlflow_client.create_model_version(
        artifact_version_id=version_id,
        description=description,
        artifact_metadata=metadata,
        internal_metadata=internal_metadata.dict(),
        data_path=INTERNAL_METADATA_PATH,
        step=step if run else None,
    )

    # update model schema at end
    update_args = {
        "version_id": version_id,
        "model_framework": framework.value,
    }
    if model_schema:
        update_args["model_schema"] = model_schema

    try:
        model_version = mlflow_client.update_model_version(**update_args)
        if model_schema:
            model_version = mlflow_client.add_custom_metrics_to_model_version(
                version_id=version_id, custom_metrics=custom_metrics
            )
    except Exception:
        # TODO (chiragjn): what is the best exception to catch here?
        logger.error(MODEL_SCHEMA_UPDATE_FAILURE_HELP.format(fqn=model_version.fqn))

    return ModelVersion.from_fqn(fqn=model_version.fqn)



import os
import typing

from mlfoundry.exceptions import MlFoundryException


class _InternalNamespace:
    NAMESPACE = "mlf"
    DELIMETER = "."
    NAMESPACE_VIOLATION_MESSAGE = """
        {name} cannot start with {prefix}
    """

    def __call__(self, name: str, delimeter: str = DELIMETER):
        if not name:
            raise MlFoundryException("name should be a non empty string")
        return _InternalNamespace.NAMESPACE + delimeter + name

    def __truediv__(self, path: str):
        return os.path.join(_InternalNamespace.NAMESPACE, path)

    @staticmethod
    def _validate_name_not_using_namespace(name: typing.Optional[str], delimeter: str):
        if name and name.startswith(_InternalNamespace.NAMESPACE + delimeter):
            raise MlFoundryException(
                _InternalNamespace.NAMESPACE_VIOLATION_MESSAGE.format(
                    name=name, prefix=_InternalNamespace.NAMESPACE + delimeter
                )
            )

    def validate_namespace_not_used(
        self,
        names: typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,
        delimeter: str = DELIMETER,
        path: typing.Optional[str] = None,
    ):
        if isinstance(names, str):
            names = [names]
        if names is not None:
            for name_ in names:
                self._validate_name_not_using_namespace(name_, delimeter)
        if path:
            prefix = os.path.normpath(os.path.join(_InternalNamespace.NAMESPACE, ""))
            if os.path.normpath(path).startswith(prefix):
                raise MlFoundryException(
                    _InternalNamespace.NAMESPACE_VIOLATION_MESSAGE.format(
                        name=path, prefix=prefix
                    )
                )


NAMESPACE = _InternalNamespace()



import enum
from typing import Dict, List, Optional

from mlfoundry.pydantic_v1 import BaseModel, constr, root_validator


class Group(enum.Enum):
    ACTUALS = "actuals"
    PREDICTIONS = "predictions"


class Position(BaseModel):
    min_x: float
    min_y: float
    max_x: float
    max_y: float

    class Config:
        allow_mutation = False


class BoundingBox(BaseModel):
    position: Position
    class_name: constr(min_length=1)
    caption: Optional[str] = None

    class Config:
        allow_mutation = False

    @root_validator(pre=True)
    def set_caption_if_not_passed(cls, values):
        if not values.get("caption"):
            values["caption"] = values.get("class_name")

        return values


class BoundingBoxGroups(BaseModel):
    __root__: Dict[Group, List[BoundingBox]]

    class Config:
        allow_mutation = False
        use_enum_values = True

    def to_dict(self):
        return self.dict()["__root__"]


class ClassGroups(BaseModel):
    __root__: Dict[Group, List[constr(min_length=1)]]

    class Config:
        allow_mutation = False
        use_enum_values = True

    @root_validator(pre=True)
    def transform_class_to_classes(cls, values):
        values = values["__root__"]
        processed_values = {}
        for group, classes in values.items():
            if isinstance(classes, str):
                processed_values[group] = [classes]
            else:
                processed_values[group] = classes
        return {"__root__": processed_values}

    def to_dict(self):
        return self.dict()["__root__"]



import mlflow

from mlfoundry.frameworks.base_registry import BaseRegistry


class SpacyRegistry(BaseRegistry):
    def log_model(self, model, artifact_path: str, **kwargs):
        """
        Log Spacy model
        Args:
            model : the model to be registered
            artifact_path (str): artifact_path
        """
        mlflow.spacy.log_model(model, artifact_path=artifact_path)

    def save_model(self, model, path: str, **kwargs):
        return mlflow.spacy.save_model(model, path, **kwargs)

    def load_model(self, model_file_path: str, dest_path: str = None, **kwargs):
        """
        Load Spacy model
        """
        return mlflow.spacy.load_model(model_file_path, dest_path)



import time

from mlflow.utils.rest_utils import MlflowHostCreds, http_request, http_request_safe

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger
from mlfoundry.tracking.entities import AuthServerInfo, DeviceCode, Token


class AuthService:
    def __init__(self, auth_server_info: AuthServerInfo):
        self._tenant_name = auth_server_info.tenant_name
        self.host_creds = MlflowHostCreds(host=auth_server_info.auth_server_url)

    def refresh_token(self, token: Token) -> Token:
        if not token.refresh_token:
            # TODO: Add a way to propagate error messages without traceback to the output interface side
            raise MlFoundryException(
                "Unable to resume login session. Please log in again using `mlfoundry login [--host HOST] --relogin`"
            )
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/api/v1/oauth/token/refresh",
            method="post",
            json={
                "tenantName": token.tenant_name,
                "refreshToken": token.refresh_token,
            },
            timeout=3,
            max_retries=0,
        )
        response = response.json()
        return Token.parse_obj(response)

    def get_device_code(self) -> DeviceCode:
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/api/v1/oauth/device",
            method="post",
            json={"tenantName": self._tenant_name},
            timeout=3,
            max_retries=0,
        )
        response = response.json()
        return DeviceCode.parse_obj(response)

    def get_token_from_device_code(
        self, device_code: str, timeout: float = 60
    ) -> Token:
        start_time = time.monotonic()
        while (time.monotonic() - start_time) <= timeout:
            response = http_request(
                host_creds=self.host_creds,
                endpoint="/api/v1/oauth/device/token",
                method="post",
                json={"tenantName": self._tenant_name, "deviceCode": device_code},
                timeout=3,
                max_retries=0,
            )
            if response.status_code == 202:
                logger.debug("User has not authorized yet. Checking again.")
                time.sleep(1.0)
                continue
            if response.status_code == 201:
                response = response.json()
                return Token.parse_obj(response)
            raise MlFoundryException(
                "Failed to get token using device code. "
                f"status_code {response.status_code},\n {response.text}"
            )
        raise MlFoundryException(f"Did not get authorized within {timeout} seconds.")



import functools
import os
import platform
import re
import time
import warnings
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urljoin, urlsplit

import mlflow
from mlflow.entities import (
    ArtifactType,
    CustomMetric,
    Metric,
    ModelSchema,
    Param,
    Run,
    RunData,
    RunInfo,
    RunStatus,
    RunTag,
)
from mlflow.tracking import MlflowClient

from mlfoundry import constants, enums, version
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.log_types import Image, Plot
from mlfoundry.log_types.artifacts.artifact import ArtifactPath, ArtifactVersion
from mlfoundry.log_types.artifacts.general_artifact import _log_artifact_version
from mlfoundry.log_types.artifacts.model import ModelVersion, _log_model_version
from mlfoundry.logger import logger
from mlfoundry.metrics.v2 import ComputedMetrics
from mlfoundry.metrics.v2 import get_metrics_calculator as get_metrics_calculator_v2
from mlfoundry.run_utils import (
    ParamsType,
    flatten_dict,
    log_artifact_blob,
    process_params,
)
from mlfoundry.session import ACTIVE_RUNS


def _ensure_not_deleted(method):
    @functools.wraps(method)
    def _check_deleted_or_not(self, *args, **kwargs):
        if self._deleted:
            raise MlFoundryException(f"Run was deleted, cannot access a deleted Run")
        else:
            return method(self, *args, **kwargs)

    return _check_deleted_or_not


class MlFoundryRun:
    """MlFoundryRun."""

    VALID_PARAM_AND_METRIC_NAMES = re.compile(r"^[A-Za-z0-9_\-\. /]+$")

    def __init__(
        self,
        experiment_id: str,
        run_id: str,
        auto_end: bool = False,
        **kwargs,
    ):
        """__init__.

        Args:
            experiment_id (str): experiment_id
            run_id (str): run_id
            auto_end (bool): If to end the run at garbage collection or process end (atexit)
        """
        self._experiment_id = str(experiment_id)
        self._run_id = run_id
        self._auto_end = auto_end
        # TODO (chiragjn): mlflow_client be a protected/private member
        self.mlflow_client = MlflowClient()
        self._run_info: Optional[RunInfo] = None
        self._run_data: Optional[RunData] = None
        self._deleted = False
        self._terminate_called = False
        if self._auto_end:
            ACTIVE_RUNS.add_run(self)

    @classmethod
    def _get_run_from_mlflow_run(cls, mlflow_run: Run):
        """classmethod to get MLfoundry run from mlfow_run instance"""
        run = cls(mlflow_run.info.experiment_id, mlflow_run.info.run_id)
        run._run_info = mlflow_run.info
        run._run_data = mlflow_run.data
        return run

    def _get_run_info(self) -> RunInfo:
        if self._run_info is not None:
            return self._run_info

        self._run_info = self.mlflow_client.get_run(self.run_id).info
        return self._run_info

    @property
    @_ensure_not_deleted
    def run_id(self) -> str:
        """Get run_id for the current `run`"""
        return self._run_id

    @property
    @_ensure_not_deleted
    def run_name(self) -> str:
        """Get run_name for the current `run`"""
        return self._get_run_info().name

    @property
    @_ensure_not_deleted
    def fqn(self) -> str:
        """Get fqn for the current `run`"""
        return self._get_run_info().fqn

    @property
    @_ensure_not_deleted
    def status(self) -> str:
        """Get status for the current `run`"""
        return self.mlflow_client.get_run(self.run_id).info.status

    @property
    @_ensure_not_deleted
    def ml_repo(self) -> str:
        """Get ml_repo name of which the current `run` is part of"""
        return self.mlflow_client.get_experiment(self._experiment_id).name

    @property
    @_ensure_not_deleted
    def auto_end(self) -> bool:
        """Tells whether automatic end for `run` is True or False"""
        return self._auto_end

    @_ensure_not_deleted
    def __repr__(self) -> str:
        return f"<{type(self).__name__} at 0x{id(self):x}: run={self.fqn!r}>"

    @_ensure_not_deleted
    def __enter__(self):
        return self

    def _terminate_run_if_running(self, termination_status: RunStatus):
        """_terminate_run_if_running.

        Args:
            termination_status (RunStatus): termination_status
        """
        if self._terminate_called:
            return

        # Prevent double execution for termination
        self._terminate_called = True
        ACTIVE_RUNS.remove_run(self)

        current_status = self.status
        termination_status = RunStatus.to_string(termination_status)
        try:
            # we do not need to set any termination status unless the run was in RUNNING state
            if current_status != RunStatus.to_string(RunStatus.RUNNING):
                return
            logger.info("Setting run status of %r to %r", self.fqn, termination_status)
            self.mlflow_client.set_terminated(self.run_id, termination_status)
        except Exception as e:
            logger.warning(
                f"failed to set termination status {termination_status} due to {e}"
            )
        print(f"Finished run: {self.fqn!r}, Dashboard: {self.dashboard_link}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        self._terminate_run_if_running(status)

    def __del__(self):
        if self._auto_end:
            self._terminate_run_if_running(RunStatus.FINISHED)

    @property
    @_ensure_not_deleted
    def dashboard_link(self) -> str:
        """Get Mlfoundry dashboard link for a `run`"""

        base_url = "{uri.scheme}://{uri.netloc}/".format(
            uri=urlsplit(mlflow.get_tracking_uri())
        )

        return urljoin(base_url, f"mlfoundry/{self._experiment_id}/run/{self.run_id}/")

    @_ensure_not_deleted
    def end(self, status: RunStatus = RunStatus.FINISHED):
        """End a `run`.

        This function marks the run as `FINISHED`.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )
            # ...
            # Model training code
            # ...
            run.end()
            ```

            In case the run was created using the context manager approach,
            We do not need to call this function.

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            with client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            ) as run:
                # ...
                # Model training code
                ...
            # `run` will be automatically marked as `FINISHED` or `FAILED`.
            ```
        """
        self._terminate_run_if_running(status)

    @_ensure_not_deleted
    def delete(self) -> None:
        """
        This function permanently delete the run

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            client.create_ml_repo('iris-learning')
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            run.log_params({"learning_rate": 0.001})
            run.log_metrics({"accuracy": 0.7, "loss": 0.6})

            run.delete()
            ```

            In case we try to call or access any other function of that run after deleting
            then it will through MlfoundryException

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            client.create_ml_repo('iris-learning')
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            run.log_params({"learning_rate": 0.001})
            run.log_metrics({"accuracy": 0.7, "loss": 0.6})

            run.delete()
            run.log_params({"learning_rate": 0.001})  # raises MlfoundryException
            ```
        """
        name = self.run_name
        try:
            self.mlflow_client.hard_delete_run(self.run_id)
            logger.info(f"Run {name} was deleted successfully")
            ACTIVE_RUNS.remove_run(self)
            self._deleted = True
            self._auto_end = False
        except Exception as ex:
            logger.warning(f"Failed to delete the run {name} because of {ex}")
            raise

    @_ensure_not_deleted
    def list_artifact_versions(
        self,
        artifact_type: Optional[List[ArtifactType]] = None,
    ) -> Iterator[ArtifactVersion]:
        """
        Get all the version of an artifact from a particular run to download contents or load them in memory

        Args:
            artifact_type: Type of the artifact you want

        Returns:
            Iterator[ArtifactVersion]: An iterator that yields non deleted artifact-versions
                of an artifact under a given run  sorted reverse by the version number

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            artifact_versions = run.list_artifact_versions()

            for artifact_version in artifact_versions:
                print(artifact_version)

            run.end()
            ```
        """
        page_token, done = None, False
        while not done:
            artifact_versions = self.mlflow_client.list_artifact_versions(
                run_ids=[self.run_id],
                artifact_types=artifact_type,
                page_token=page_token,
            )
            for artifact_version in artifact_versions:
                yield ArtifactVersion.from_fqn(artifact_version.fqn)
            done = page_token is None

    @_ensure_not_deleted
    def list_model_versions(
        self,
    ) -> Iterator[ModelVersion]:
        """
        Get all the version of a models from a particular run to download contents or load them in memory

        Returns:
            Iterator[ModelVersion]: An iterator that yields non deleted model-versions
                under a given run  sorted reverse by the version number

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.get_run(run_id="<your-run-id>")
            model_versions = run.list_model_versions()

            for model_version in model_versions:
                print(model_version)

            run.end()
            ```
        """
        page_token, done = None, False
        while not done:
            model_versions = self.mlflow_client.list_model_versions(
                page_token=page_token,
                run_ids=[self.run_id],
            )
            for model_version in model_versions:
                yield ModelVersion.from_fqn(fqn=model_version.model_fqn)
            done = page_token is None

    def _add_git_info(self, root_path: Optional[str] = None):
        """_add_git_info.

        Args:
            root_path (Optional[str]): root_path
        """
        root_path = root_path or os.getcwd()
        try:
            from mlfoundry.git_info import GitInfo

            git_info = GitInfo(root_path)
            tags = [
                RunTag(
                    key=constants.GIT_COMMIT_TAG_NAME,
                    value=git_info.current_commit_sha,
                ),
                RunTag(
                    key=constants.GIT_BRANCH_TAG_NAME,
                    value=git_info.current_branch_name,
                ),
                RunTag(key=constants.GIT_DIRTY_TAG_NAME, value=str(git_info.is_dirty)),
            ]
            remote_url = git_info.remote_url
            if remote_url is not None:
                tags.append(RunTag(key=constants.GIT_REMOTE_URL_NAME, value=remote_url))
            self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)
            log_artifact_blob(
                mlflow_client=self.mlflow_client,
                run_id=self.run_id,
                blob=git_info.diff_patch,
                file_name=constants.PATCH_FILE_NAME,
                artifact_path=constants.PATCH_FILE_ARTIFACT_DIR,
            )
        except Exception as ex:
            # no-blocking
            logger.warning(f"failed to log git info because {ex}")

    def _add_python_mlf_version(self):
        python_version = platform.python_version()
        mlfoundry_version = version.__version__

        tags = [
            RunTag(
                key=constants.PYTHON_VERSION_TAG_NAME,
                value=python_version,
            ),
        ]

        if mlfoundry_version:
            tags.append(
                RunTag(
                    key=constants.MLFOUNDRY_VERSION_TAG_NAME,
                    value=mlfoundry_version,
                )
            )
        else:
            logger.warning("Failed to get MLFoundry version.")

        self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)

    @_ensure_not_deleted
    def _download_artifact_deprecated(
        self, path: str, dest_path: Optional[str] = None
    ) -> str:
        """Downloads a logged `artifact` associated with the current `run`.

        Args:
            path (str): Source artifact path.
            dest_path (Optional[str], optional): Absolute path of the local destination
                directory. If a directory path is passed, the directory must already exist.
                If not passed, the artifact will be downloaded to a newly created directory
                in the local filesystem.

        Returns:
            str: Path of the directory where the artifact is downloaded.

        Examples:

            ```python
            import os
            import mlfoundry

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )

            run.log_artifact(local_path="artifact.txt", artifact_path="my-artifacts")

            local_path = run.download_artifact(path="my-artifacts")
            print(f"Artifacts: {os.listdir(local_path)}")

            run.end()
            ```
        """
        if dest_path is None:
            return self.mlflow_client.download_artifacts(self.run_id, path=path)
        elif os.path.isdir(dest_path):
            return self.mlflow_client.download_artifacts(
                self.run_id, path=path, dst_path=dest_path
            )
        else:
            raise MlFoundryException(
                f"Destination path {dest_path} should be an existing directory."
            )

    @_ensure_not_deleted
    def _log_artifact_deprecated(
        self, local_path: str, artifact_path: Optional[str] = None
    ):
        """Logs an artifact for the current `run`.

        An `artifact` is a local file or directory. This function stores the `artifact`
        at the remote artifact storage.

        Args:
            local_path (str): Local path of the file or directory.
            artifact_path (Optional[str], optional): Relative destination path where
                the `artifact` will be stored. If not passed, the `artifact` is stored
                at the root path of the `run`'s artifact storage. The passed
                `artifact_path` should not start with `mlf/`, as `mlf/` directory is
                reserved for `mlfoundry`.

        Examples:

            ```python
            import os
            import mlfoundry

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )

            run.log_artifact(local_path="artifact.txt", artifact_path="my-artifacts")

            run.end()
            ```
        """
        # TODO (chiragjn): this api is a little bit confusing, artifact_path is always considered to be a directory
        # which means passing local_path="a/b/c/d.txt", artifact_path="x/y/z.txt" will result in x/y/z.txt/d.txt

        if artifact_path is not None:
            NAMESPACE.validate_namespace_not_used(path=artifact_path)
        if os.path.isfile(local_path):
            _to = os.path.join(artifact_path or "", os.path.basename(local_path))
            logger.info(
                f"Logging {local_path!r} file as artifact to {_to!r}, this might take a while ..."
            )
            self.mlflow_client.log_artifact(
                self.run_id, local_path=local_path, artifact_path=artifact_path
            )
        elif os.path.isdir(local_path):
            _to = artifact_path or "/"
            logger.info(
                f"Logging contents of the {local_path!r} directory as artifacts to {_to!r}, this might take a while ..."
            )
            self.mlflow_client.log_artifacts(
                self.run_id, local_dir=local_path, artifact_path=artifact_path
            )
        else:
            raise MlFoundryException(
                f"local path {local_path} should be an existing file or directory"
            )

    @_ensure_not_deleted
    def log_artifact(
        self,
        name: str,
        artifact_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], ArtifactPath]
        ],
        description: Optional[str] = None,
        metadata: Optional[Dict] = None,
        step: Optional[int] = 0,
        progress: Optional[bool] = None,
    ) -> ArtifactVersion:
        """Logs an artifact for the current ML Repo.

        An `artifact` is a list of local files and directories.
        This function packs the mentioned files and directories in `artifact_paths`
        and uploads them to remote storage linked to the experiment

        Args:
            name (str): Name of the Artifact. If an artifact with this name already exists under the current ML Repo,
                the logged artifact will be added as a new version under that `name`.
                If no artifact exist with the given `name`, the given artifact will be logged as version 1.
            artifact_paths (List[mlfoundry.ArtifactPath], optional): A list of pairs
                of (source path, destination path) to add files and folders
                to the artifact version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the artifact contents to upload to.

                ```
                E.g. >>> run.log_artifact(
                     ...     name="xyz",
                     ...     artifact_paths=[
                                mlfoundry.ArtifactPath("foo.txt", "foo/bar/foo.txt"),
                                mlfoundry.ArtifactPath("tokenizer/", "foo/tokenizer/"),
                                mlfoundry.ArtifactPath('bar.text'),
                                ('bar.txt', ),
                                ('foo.txt', 'a/foo.txt')
                             ]
                     ... )
                would result in
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            step (int): step/iteration at which the vesion is being logged, defaults to 0.
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            mlfoundry.ArtifactVersion: an instance of `ArtifactVersion` that can be used to download the files,
            or update attributes like description, metadata.

        Examples:

            ```python
            import os
            import mlfoundry

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )

            run.log_artifact(
                name="hello-world-file",
                artifact_paths=[mlfoundry.ArtifactPath('artifact.txt', 'a/b/')]
            )

            run.end()
            ```
        """
        if not artifact_paths:
            raise MlFoundryException(
                "artifact_paths cannot be empty, atleast one artifact_path must be passed"
            )

        return _log_artifact_version(
            self,
            name=name,
            artifact_paths=artifact_paths,
            description=description,
            metadata=metadata,
            step=step,
            progress=progress,
        )

    @_ensure_not_deleted
    def log_metrics(self, metric_dict: Dict[str, Union[int, float]], step: int = 0):
        """Log metrics for the current `run`.

        A metric is defined by a metric name (such as "training-loss") and a
        floating point or integral value (such as `1.2`). A metric is associated
        with a `step` which is the training iteration at which the metric was
        calculated.

        Args:
            metric_dict (Dict[str, Union[int, float]]): A metric name to metric value map.
                metric value should be either `float` or `int`. This should be
                a non-empty dictionary.
            step (int, optional): Training step/iteration at which the metrics
                present in `metric_dict` were calculated. If not passed, `0` is
                set as the `step`.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
            run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

            run.end()
            ```
        """
        try:
            # mlfow_client doesn't have log_metrics api, so we have to use log_batch,
            # This is what internally used by mlflow.log_metrics
            timestamp = int(time.time() * 1000)
            metrics_arr = []
            for key in metric_dict.keys():
                if isinstance(metric_dict[key], str):
                    logger.warning(
                        f"Cannot log metric with string value. Discarding metric {key}={metric_dict[key]}"
                    )
                    continue
                if not self.VALID_PARAM_AND_METRIC_NAMES.match(key):
                    logger.warning(
                        f"Invalid metric name: {key}. Names may only contain alphanumerics, "
                        f"underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                        f"Discarding metric {key}={metric_dict[key]}"
                    )
                    continue
                metrics_arr.append(Metric(key, metric_dict[key], timestamp, step=step))

            if len(metrics_arr) == 0:
                logger.warning("Cannot log empty metrics dictionary")
                return

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=metrics_arr, params=[], tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        logger.info("Metrics logged successfully")

    @_ensure_not_deleted
    def log_params(self, param_dict: ParamsType, flatten_params: bool = False):
        """Logs parameters for the run.

        Parameters or Hyperparameters can be thought of as configurations for a run.
        For example, the type of kernel used in a SVM model is a parameter.
        A Parameter is defined by a name and a string value. Parameters are
        also immutable, we cannot overwrite parameter value for a parameter
        name.

        Args:
            param_dict (ParamsType): A parameter name to parameter value map.
                Parameter values are converted to `str`.
            flatten_params (bool): Flatten hierarchical dict, e.g. `{'a': {'b': 'c'}} -> {'a.b': 'c'}`.
                All the keys will be converted to `str`. Defaults to False

        Examples:

            ### Logging parameters using a `dict`.

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params({"learning_rate": 0.01, "epochs": 10})

            run.end()
            ```

            ### Logging parameters using `argparse` Namespace object

            ```python
            import argparse
            import mlfoundry

            parser = argparse.ArgumentParser()
            parser.add_argument("-batch_size", type=int, required=True)
            args = parser.parse_args()

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params(args)
            ```
        """
        try:
            # mlfow client doesn't have log_params api, so we have to use log_batch,
            # This is what internally used by mlflow.log_params
            param_dict = process_params(param_dict)
            param_dict = flatten_dict(param_dict) if flatten_params else param_dict

            params_arr = []
            for param_key in param_dict.keys():
                if len(str(param_key)) > 250 or len(str(param_dict[param_key])) > 250:
                    logger.warning(
                        f"MlFoundry can't log parmeters with length greater than 250 characters. "
                        f"Discarding {param_key}:{param_dict[param_key]}."
                    )
                    continue
                if not self.VALID_PARAM_AND_METRIC_NAMES.match(param_key):
                    logger.warning(
                        f"Invalid param name: {param_key}. Names may only contain alphanumerics, "
                        f"underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                        f"Discarding param {param_key}={param_dict[param_key]}"
                    )
                    continue
                params_arr.append(Param(param_key, str(param_dict[param_key])))

            if len(params_arr) == 0:
                logger.warning("Cannot log empty params dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=params_arr, tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None
        logger.info("Parameters logged successfully")

    @_ensure_not_deleted
    def set_tags(self, tags: Dict[str, str]):
        """Set tags for the current `run`.

        Tags are "labels" for a run. A tag is represented by a tag name and value.

        Args:
            tags (Dict[str, str]): A tag name to value map.
                Tag name cannot start with `mlf.`, `mlf.` prefix
                is reserved for mlfoundry. Tag values will be converted
                to `str`.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.set_tags({"nlp.framework": "Spark NLP"})

            run.end()
            ```
        """
        try:
            NAMESPACE.validate_namespace_not_used(names=tags.keys())
            tags_arr = [RunTag(key, str(value)) for key, value in tags.items()]
            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=[], tags=tags_arr
            )
        except MlflowException as e:
            raise MlFoundryException(e.message) from e
        logger.info("Tags set successfully")

    @_ensure_not_deleted
    def get_tags(self, no_cache=False) -> Dict[str, str]:
        """Returns all the tags set for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing tags. The keys in the dictionary
                are tag names and the values are corresponding tag values.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.set_tags({"nlp.framework": "Spark NLP"})
            print(run.get_tags())

            run.end()
            ```
        """
        if not no_cache and self._run_data:
            return self._run_data.tags

        run = self.mlflow_client.get_run(self.run_id)
        self._run_data = run.data
        return self._run_data.tags

    @_ensure_not_deleted
    def auto_log_metrics(
        self,
        model_type: enums.ModelType,
        data_slice: enums.DataSlice,
        predictions: Collection[Any],
        actuals: Optional[Collection[Any]] = None,
        class_names: Optional[List[str]] = None,
        prediction_probabilities=None,
    ) -> ComputedMetrics:
        """auto_log_metrics.

        Args:
            model_type (enums.ModelType): model_type
            data_slice (enums.DataSlice): data_slice
            predictions (Collection[Any]): predictions
            actuals (Optional[Collection[Any]]): actuals
            class_names (Optional[List[str]]): class_names
            prediction_probabilities:

        Returns:
            ComputedMetrics:
        """
        metrics_calculator = get_metrics_calculator_v2(model_type)
        metrics = metrics_calculator.compute_metrics(
            predictions=predictions,
            actuals=actuals,
            prediction_probabilities=prediction_probabilities,
            class_names=class_names,
        )
        metric_path = os.path.join(constants.ALM_ARTIFACT_PATH, data_slice.value)
        log_artifact_blob(
            mlflow_client=self.mlflow_client,
            run_id=self.run_id,
            blob=metrics.json(),
            file_name=constants.ALM_METRICS_FILE_NAME,
            artifact_path=metric_path,
        )
        return metrics

    @_ensure_not_deleted
    def get_metrics(
        self, metric_names: Optional[Iterable[str]] = None
    ) -> Dict[str, List[Metric]]:
        """Get metrics logged for the current `run` grouped by metric name.

        Args:
            metric_names (Optional[Iterable[str]], optional): A list of metric names
                For which the logged metrics will be fetched. If not passed, then all
                metrics logged under the `run` is returned.

        Returns:
            Dict[str, List[Metric]]: A dictionary containing metric name to list of metrics
                map.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )
            run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
            run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

            metrics = run.get_metrics()
            for metric_name, metric_history in metrics.items():
                print(f"logged metrics for metric {metric_name}:")
                for metric in metric_history:
                    print(f"value: {metric.value}")
                    print(f"step: {metric.step}")
                    print(f"timestamp_ms: {metric.timestamp}")
                    print("--")

            run.end()
            ```
        """
        run = self.mlflow_client.get_run(self.run_id)
        run_metrics = run.data.metrics

        metric_names = (
            set(metric_names) if metric_names is not None else run_metrics.keys()
        )

        unknown_metrics = metric_names - run_metrics.keys()
        if len(unknown_metrics) > 0:
            logger.warning(f"{unknown_metrics} metrics not present in the run")
        metrics_dict = {metric_name: [] for metric_name in unknown_metrics}
        valid_metrics = metric_names - unknown_metrics
        for metric_name in valid_metrics:
            metrics_dict[metric_name] = self.mlflow_client.get_metric_history(
                self.run_id, metric_name
            )
        return metrics_dict

    @_ensure_not_deleted
    def get_params(self, no_cache=False) -> Dict[str, str]:
        """Get all the params logged for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing the parameters. The keys in the dictionary
                are parameter names and the values are corresponding parameter values.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params({"learning_rate": 0.01, "epochs": 10})
            print(run.get_params())

            run.end()
            ```
        """
        if not no_cache and self._run_data:
            return self._run_data.params

        run = self.mlflow_client.get_run(self.run_id)
        self._run_data = run.data
        return self._run_data.params

    @_ensure_not_deleted
    def log_model(
        self,
        *,
        name: str,
        model_file_or_folder: Optional[str] = None,
        framework: Optional[Union[enums.ModelFramework, str]],
        additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]] = (),
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        model_schema: Optional[Union[Dict[str, Any], ModelSchema]] = None,
        custom_metrics: Optional[List[Union[Dict[str, Any], CustomMetric]]] = None,
        step: int = 0,
        progress: Optional[bool] = None,
        # Deprecated
        model: Optional[Any] = None,
        model_save_kwargs: Optional[Dict[str, Any]] = None,
    ) -> ModelVersion:
        # TODO (chiragjn): Document mapping of framework to list of valid model save kwargs
        # TODO (chiragjn): Add more examples
        """
        Serialize and log a versioned model under the current ML Repo. Each logged model generates a new version
            associated with the given `name` and linked to the current run. Multiple versions of the model can be
            logged as separate versions under the same `name`.

        Args:
            name (str): Name of the model. If a model with this name already exists under the current ML Repo,
                the logged model will be added as a new version under that `name`. If no models exist with the given
                `name`, the given model will be logged as version 1.
            model_file_or_folder (str): Path to either a single file or a folder containing model files. This folder
                is usually created using serialization methods of libraries or frameworks e.g. `joblib.dump`,
                `model.save_pretrained(...)`, `torch.save(...)`, `model.save(...)`
            framework (Union[enums.ModelFramework, str]): Model Framework. Ex:- pytorch, sklearn, tensorflow etc.
                The full list of supported frameworks can be found in `mlfoundry.enums.ModelFramework`.
                Can also be `None` when `model` is `None`.
            additional_files (Sequence[Tuple[Union[str, Path], Optional[str]]], optional): A list of pairs
                of (source path, destination path) to add additional files and folders
                to the model version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the model versions contents to upload to.
                The model version contents are arranged like follows
                .
                └── model/
                    └── # model files are serialized here
                └── # any additional files and folders can be added here.

                You can also add additional files to model/ subdirectory by specifying the destination path as model/

                ```
                E.g. >>> run.log_model(
                     ...     name="xyz", model_file_or_folder="clf.joblib", framework="sklearn",
                     ...     additional_files=[("foo.txt", "foo/bar/foo.txt"), ("tokenizer/", "foo/tokenizer/")]
                     ... )
                would result in
                .
                ├── model/
                │   └── clf.joblib  # if `model_file_or_folder` is a folder, contents will be added here
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            model_schema (Optional[Union[Dict[str, Any], ModelSchema]], optional): instance of `mlfoundry.ModelSchema`.
                This schema needs to be consistent with older versions of the model under the given `name` i.e.
                a feature's value type and model's prediction type cannot be changed in the schema of new version.
                Features can be removed or added between versions.
                ```
                E.g. if there exists a v1 with
                    schema = {"features": {"name": "feat1": "int"}, "prediction": "categorical"}, then

                    schema = {"features": {"name": "feat1": "string"}, "prediction": "categorical"} or
                    schema = {"features": {"name": "feat1": "int"}, "prediction": "numerical"}
                    are invalid because they change the types of existing features and prediction

                    while
                    schema = {"features": {"name": "feat1": "int", "feat2": "string"}, "prediction": "categorical"} or
                    schema = {"features": {"feat2": "string"}, "prediction": "categorical"}
                    are valid

                    This field can be updated at any time after logging. Defaults to None
                ```
            custom_metrics: (Optional[Union[List[Dict[str, Any]], CustomMetric]], optional): list of instances of
                `mlfoundry.CustomMetric`
                The custom metrics must be added according to the prediction type of schema.
                custom_metrics = [{
                    "name": "mean_square_error",
                    "type": "metric",
                    "value_type": "float"
                }]
            step (int): step/iteration at which the model is being logged, defaults to 0.
            progress (bool): value to show progress bar, defaults to None.
            model (Any): Deprecated. model instance of any one of the supported frameworks
                under `mlfoundry.enums.ModelFramework`. Can also be `None` which can be useful to
                create a reference entry without uploading any model files.
            model_save_kwargs (Optional[Dict[str, Any]], optional): Deprecated. keyword arguments to pass
                to model serializer.Defaults to `None`

        Returns:
            mlfoundry.ModelVersion: an instance of `ModelVersion` that can be used to download the files,
                load the model, or update attributes like description, metadata, schema.

        Examples:

            ### Sklearn

            ```python
            import mlfoundry
            from mlfoundry.enums import ModelFramework

            import joblib
            import numpy as np
            from sklearn.pipeline import make_pipeline
            from sklearn.preprocessing import StandardScaler
            from sklearn.svm import SVC

            X = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
            y = np.array([1, 1, 2, 2])
            clf = make_pipeline(StandardScaler(), SVC(gamma='auto'))
            clf.fit(X, y)
            joblib.dump(clf, "sklearn-pipeline.joblib")

            client = mlfoundry.get_client()
            client.create_ml_repo(  # This is only required once
                ml_repo="my-classification-project",
                # This controls which bucket is used.
                # You can get this from Integrations > Blob Storage. `None` picks the default
                storage_integration_fqn=None
            )
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            model_version = run.log_model(
                name="my-sklearn-model",
                model_file_or_folder="sklearn-pipeline.joblib",
                framework=ModelFramework.SKLEARN,
                metadata={"accuracy": 0.99, "f1": 0.80},
                step=1,  # step number, useful when using iterative algorithms like SGD
            )
            print(model_version.fqn)
            ```

            ### Huggingface Transformers

            ```python
            import mlfoundry
            from mlfoundry.enums import ModelFramework

            import torch
            from transformers import AutoTokenizer, AutoConfig, pipeline, AutoModelForCausalLM
            pln = pipeline(
                "text-generation",
                model_file_or_folder="EleutherAI/pythia-70m",
                tokenizer="EleutherAI/pythia-70m",
                torch_dtype=torch.float16
            )
            pln.model.save_pretrained("my-transformers-model")
            pln.tokenizer.save_pretrained("my-transformers-model")

            client = mlfoundry.get_client()
            client.create_ml_repo(  # This is only required once
                ml_repo="my-llm-project",
                # This controls which bucket is used.
                # You can get this from Integrations > Blob Storage. `None` picks the default
                storage_integration_fqn=None
            )
            run = client.create_run(
                ml_repo="my-llm-project"
            )
            model_version = run.log_model(
                name="my-transformers-model",
                model_file_or_folder="my-transformers-model/",
                framework=ModelFramework.TRANSFORMERS
            )
            print(model_version.fqn)
            ```

        """
        if model or model_save_kwargs:
            deprecation_message = (
                "`model` and `model_save_kwargs` are deprecated and will be removed in a future release. "
                "Please use `model_file_or_folder` argument instead to pass an already serialized model"
            )
            warnings.warn(
                message=deprecation_message, category=DeprecationWarning, stacklevel=2
            )
            logger.warning(deprecation_message)
        model_version = _log_model_version(
            run=self,
            name=name,
            model_file_or_folder=model_file_or_folder,
            framework=framework,
            additional_files=additional_files,
            description=description,
            metadata=metadata,
            model_schema=model_schema,
            custom_metrics=custom_metrics,
            step=step,
            progress=progress,
            # Deprecated:
            model=model,
            model_save_kwargs=model_save_kwargs,
        )
        logger.info(f"Logged model successfully with fqn {model_version.fqn!r}")
        return model_version

    @_ensure_not_deleted
    def log_images(self, images: Dict[str, Image], step: int = 0):
        """Log images under the current `run` at the given `step`.

        Use this function to log images for a `run`. `PIL` package is needed to log images.
        To install the `PIL` package, run `pip install pillow`.

        Args:
            images (Dict[str, "mlfoundry.Image"]): A map of string image key to instance of
                `mlfoundry.Image` class. The image key should only contain alphanumeric,
                hyphens(-) or underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `images` should be
                logged. Default is `0`.

        Examples:

            ### Logging images from different sources

            ```python
            import mlfoundry
            import numpy as np
            import PIL.Image

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))
            im = PIL.Image.fromarray(imarray.astype("uint8")).convert("RGB")
            im.save("result_image.jpeg")

            images_to_log = {
                "logged-image-array": mlfoundry.Image(data_or_path=imarray),
                "logged-pil-image": mlfoundry.Image(data_or_path=im),
                "logged-image-from-path": mlfoundry.Image(data_or_path="result_image.jpeg"),
            }

            run.log_images(images_to_log, step=1)
            run.end()
            ```
        """
        for key, image in images.items():
            if not isinstance(image, Image):
                raise MlFoundryException("image should be of type `mlfoundry.Image`")
            image.save(run=self, key=key, step=step)

    @_ensure_not_deleted
    def log_plots(
        self,
        plots: Dict[
            str,
            Union[
                "matplotlib.pyplot",
                "matplotlib.figure.Figure",
                "plotly.graph_objects.Figure",
                Plot,
            ],
        ],
        step: int = 0,
    ):
        """Log custom plots under the current `run` at the given `step`.

        Use this function to log custom matplotlib, plotly plots.

        Args:
            plots (Dict[str, "matplotlib.pyplot", "matplotlib.figure.Figure", "plotly.graph_objects.Figure", Plot]):
                A map of string plot key to the plot or figure object.
                The plot key should only contain alphanumeric, hyphens(-) or
                underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `plots` should be
                logged. Default is `0`.


        Examples:

            ### Logging a plotly figure

            ```python
            import mlfoundry
            import plotly.express as px

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            df = px.data.tips()
            fig = px.histogram(
                df,
                x="total_bill",
                y="tip",
                color="sex",
                marginal="rug",
                hover_data=df.columns,
            )

            plots_to_log = {
                "distribution-plot": fig,
            }

            run.log_plots(plots_to_log, step=1)
            run.end()
            ```


            ### Logging a matplotlib plt or figure

            ```python
            import mlfoundry
            from matplotlib import pyplot as plt
            import numpy as np

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            t = np.arange(0.0, 5.0, 0.01)
            s = np.cos(2 * np.pi * t)
            (line,) = plt.plot(t, s, lw=2)

            plt.annotate(
                "local max",
                xy=(2, 1),
                xytext=(3, 1.5),
                arrowprops=dict(facecolor="black", shrink=0.05),
            )

            plt.ylim(-2, 2)

            plots_to_log = {"cos-plot": plt, "cos-plot-using-figure": plt.gcf()}

            run.log_plots(plots_to_log, step=1)
            run.end()
            ```
        """
        for key, plot in plots.items():
            plot = Plot(plot) if not isinstance(plot, Plot) else plot
            plot.save(run=self, key=key, step=step)



from __future__ import annotations

import os
import threading
from functools import lru_cache, wraps
from pathlib import Path
from typing import Optional

import click
from filelock import FileLock, Timeout

from mlfoundry.env_vars import API_KEY_GLOBAL, TRACKING_HOST_GLOBAL
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger
from mlfoundry.pydantic_v1 import BaseModel, Field, constr
from mlfoundry.run_utils import resolve_tracking_uri
from mlfoundry.tracking.entities import Token
from mlfoundry.tracking.servicefoundry_service import ServicefoundryService
from mlfoundry.tracking.truefoundry_rest_store import get_rest_store

CREDENTIALS_DIR = Path.home() / ".truefoundry"
CREDENTIALS_FILE = CREDENTIALS_DIR / "credentials.json"


OLD_CREDENTIALS_DIR = Path.home() / ".mlfoundry"
OLD_CREDENTIALS_FILE = OLD_CREDENTIALS_DIR / "credentials.netrc"

if OLD_CREDENTIALS_FILE.exists():
    logger.warning(
        "%s file is deprecated. You can delete this file now.", OLD_CREDENTIALS_FILE
    )


class CredentialsFileContent(BaseModel):
    access_token: constr(min_length=1) = Field(repr=False)
    refresh_token: Optional[constr(min_length=1)] = Field(repr=False)
    host: constr(min_length=1)

    class Config:
        allow_mutation = False

    def to_token(self) -> Token:
        return Token(access_token=self.access_token, refresh_token=self.refresh_token)


def _ensure_lock_taken(method):
    @wraps(method)
    def lock_guard(self, *method_args, **method_kwargs):
        if not self.lock_taken():
            raise Exception(
                "Trying to write to credential file without using with block"
            )
        return method(self, *method_args, **method_kwargs)

    return lock_guard


CRED_FILE_THREAD_LOCK = threading.RLock()


@lru_cache(maxsize=None)
def get_file_lock(lock_file_path: str) -> FileLock:
    return FileLock(lock_file_path)


class CredentialsFileManager:
    def __init__(
        self,
        credentials_file_path: Path = CREDENTIALS_FILE,
        lock_timeout: float = 60.0,
    ):
        credentials_file_path = credentials_file_path.absolute()

        logger.debug("credential file path %r", credentials_file_path)

        credentials_lock_file_path = f"{credentials_file_path}.lock"

        logger.debug("credential lock file path %r", credentials_lock_file_path)
        self._credentials_file_path = credentials_file_path

        cred_file_dir = credentials_file_path.parent
        cred_file_dir.mkdir(exist_ok=True, parents=True)

        self._file_lock = get_file_lock(credentials_lock_file_path)
        self._lock_timeout = lock_timeout
        self._lock_owner: Optional[int] = None

    def __enter__(self) -> CredentialsFileManager:
        # The lock objects are recursive locks, which means that once acquired, they will not block on successive lock requests:
        lock_aquired = CRED_FILE_THREAD_LOCK.acquire(timeout=self._lock_timeout)
        if not lock_aquired:
            raise MlFoundryException(
                "Could not aquire CRED_FILE_THREAD_LOCK"
                f" in {self._lock_timeout} seconds"
            )
        try:
            self._file_lock.acquire(timeout=self._lock_timeout)
        except Timeout as ex:
            raise MlFoundryException(
                f"Failed to aquire lock on credential file within {self._lock_timeout} seconds.\n"
                "Is any other process trying to login?"
            ) from ex
        logger.debug("Acquired file and thread lock to access credential file")
        self._lock_owner = threading.get_ident()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._file_lock.release()
        CRED_FILE_THREAD_LOCK.release()
        logger.debug("Released file and thread lock to access credential file")
        self._lock_owner = None

    def lock_taken(self) -> bool:
        return self._lock_owner == threading.get_ident()

    @_ensure_lock_taken
    def read(self) -> CredentialsFileContent:
        try:
            return CredentialsFileContent.parse_file(self._credentials_file_path)
        except Exception as ex:
            raise MlFoundryException(
                "Error while reading the credentials file "
                f"{self._credentials_file_path}. Please login again "
                "using `mlfoundry login` command or `mlfoundry.login()` function"
            ) from ex

    @_ensure_lock_taken
    def write(self, credentials_file_content: CredentialsFileContent):
        if not isinstance(credentials_file_content, CredentialsFileContent):
            raise MlFoundryException(
                "Only object of type `CredentialsFileContent` is allowed. "
                f"Got {type(credentials_file_content)}"
            )
        logger.debug("Updating the credential file content")
        with open(self._credentials_file_path, "w", encoding="utf8") as file:
            file.write(credentials_file_content.json())

    @_ensure_lock_taken
    def exists(self) -> bool:
        return self._credentials_file_path.exists()


def login(
    tracking_uri: Optional[str] = None,
    relogin: bool = False,
    api_key: Optional[str] = None,
) -> bool:
    """Save API key in local file system for a given `tracking_uri`.

    Args:
        tracking_uri (Optional[str], optional): tracking_uri for the given API key
        relogin (bool, optional): Overwrites the existing API key for the `tracking_uri` if
            set to `True`. If set to `False` and an API key is already present for
            the given `tracking_uri`, then the existing API key is kept untouched.
            Default is `False`.
        api_key (Optional[str], optional): The API key for the given `tracking_uri`.
            If `api_key` is not passed, this function prompts for the API key.

    Returns:
        bool: Returns `True` if any credential was persisted.
    """
    from mlfoundry.session import EnvCredentialProvider

    if API_KEY_GLOBAL in os.environ and TRACKING_HOST_GLOBAL in os.environ:
        logger.warning(
            "Skipping login because environment variables %s and "
            "%s are set and will be used when running mlfoundry. "
            "If you want to relogin then unset these environment keys.",
            TRACKING_HOST_GLOBAL,
            API_KEY_GLOBAL,
        )
        return False

    if EnvCredentialProvider.can_provide():
        logger.warning(
            "TFY_API_KEY env var is already set. "
            "When running mlfoundry, it will use the api key to authorize.\n"
            "Login will just save the credentials on disk."
        )

    tracking_uri = resolve_tracking_uri(tracking_uri).strip("/")
    auth_service = get_rest_store(tracking_uri).get_auth_service()

    cred_file = CredentialsFileManager()

    with cred_file:
        if not relogin and cred_file.exists():
            cred_file_content = cred_file.read()
            if tracking_uri != cred_file_content.host:
                if click.confirm(
                    f"Already logged in to {cred_file_content.host!r}\n"
                    f"Do you want to relogin to {tracking_uri!r}?"
                ):
                    return login(
                        tracking_uri=tracking_uri, relogin=True, api_key=api_key
                    )
            user_info = cred_file_content.to_token().to_user_info()
            user_name_display_info = user_info.email or user_info.user_type.value
            print(
                f"Already logged in to {cred_file_content.host!r} as "
                f"{user_info.user_id!r} ({user_name_display_info!r})\n"
                "Please use `mlfoundry login --relogin` or `mlfoundry.login(relogin=True)` "
                "to force relogin"
            )
            return False

        if api_key:
            logger.debug("Logging in with api key")
            servicefoundry_service = ServicefoundryService(tracking_uri)
            token = servicefoundry_service.get_token_from_api_key(api_key=api_key)
            cred_file_content = CredentialsFileContent(
                access_token=token.access_token,
                refresh_token=token.refresh_token,
                host=tracking_uri,
            )
            cred_file.write(cred_file_content)
        else:
            device_code = auth_service.get_device_code()
            url_to_go = device_code.get_user_clickable_url(tracking_uri=tracking_uri)
            print(f"Opening:- {url_to_go}")
            print(
                "Please click on the above link if it is not "
                "automatically opened in a browser window."
            )
            click.launch(url_to_go)
            token = auth_service.get_token_from_device_code(
                device_code=device_code.device_code, timeout=120
            )
            cred_file_content = CredentialsFileContent(
                access_token=token.access_token,
                refresh_token=token.refresh_token,
                host=tracking_uri,
            )
            cred_file.write(cred_file_content)

    user_info = token.to_user_info()
    user_name_display_info = user_info.email or user_info.user_type.value
    print(
        f"logged in to {cred_file_content.host!r} as {user_info.user_id!r} ({user_name_display_info})"
    )
    return True



import enum
import json
import os
import posixpath
import sys
import tempfile
from typing import Union

from mlflow.entities import ArtifactType

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.artifact import (
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from mlfoundry.log_types.artifacts.constants import FILES_DIR, INTERNAL_METADATA_PATH
from mlfoundry.log_types.pydantic_base import PydanticBase
from mlfoundry.log_types.utils import validate_key_name
from mlfoundry.pydantic_v1 import BaseModel

PlotObjType = Union[
    "matplotlib.figure.Figure",
    "plotly.graph_objects.Figure",
    "matplotlib.pyplot",
]


@enum.unique
class Format(enum.Enum):
    SVG = "SVG"
    HTML = "HTML"
    PNG = "PNG"


def _is_matplotlib_figure(fig) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    import matplotlib

    return isinstance(fig, matplotlib.figure.Figure)


def _is_matplotlib_plt(plt) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    return getattr(plt, "__name__", "") == "matplotlib.pyplot"


def _is_plotly_figure(fig) -> bool:
    if "plotly" not in sys.modules:
        return False

    import plotly

    return isinstance(fig, plotly.graph_objects.Figure)


def get_plot_file_name(format: Format) -> str:
    return f"plot.{format.value.lower()}"


class PlotArtifact(BaseModel):
    artifact_file: str
    format: Format

    class Config:
        allow_mutation = False
        use_enum_values = True


def _save_matplotlib_figure(
    figure: "matplotlib.figure.Figure",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    supported_formats = figure.canvas.get_supported_filetypes().keys()
    if "svg" in supported_formats:
        format_ = Format.SVG
    elif "png" in supported_formats:
        format_ = Format.PNG
    else:
        raise MlFoundryException(
            f"Could not save {key} {figure} matplotlib figure"
            "in either SVG or PNG format"
        )
    file_path = get_plot_file_name(format=format_)
    local_path = os.path.join(local_dir, file_path)
    figure.savefig(local_path)
    return PlotArtifact(artifact_file=file_path, format=format_)


def _save_matplotlib_plt(
    plt: "matplotlib.pyplot",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    figure = plt.gcf()
    return _save_matplotlib_figure(
        figure=figure, key=key, step=step, local_dir=local_dir
    )


def _save_plotly_figure(
    figure: "plotly.graph_objects.Figure",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    format_ = Format.HTML
    file_path = get_plot_file_name(format=format_)
    local_path = os.path.join(local_dir, file_path)
    figure.write_html(local_path, include_plotlyjs="cdn", auto_open=False)
    return PlotArtifact(artifact_file=file_path, format=format_)


class Plot:
    def __init__(self, plot_obj: PlotObjType):
        self._plot_obj = plot_obj

    def _save_plot(self, key: str, step: int, local_dir: str) -> PlotArtifact:
        if _is_matplotlib_plt(self._plot_obj):
            return _save_matplotlib_plt(
                plt=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        if _is_matplotlib_figure(self._plot_obj):
            return _save_matplotlib_figure(
                figure=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        if _is_plotly_figure(self._plot_obj):
            return _save_plotly_figure(
                figure=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        raise MlFoundryException(
            f"Unknown type: {type(self._plot_obj)}"
            "Supported types are, matplotlib.figure.Figure, matplotlib.pyplot"
            " and plotly.graph_objects.Figure"
        )

    def save(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
    ):
        validate_key_name(key)

        # creating a temp dir which will be logged
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")
        local_files_dir = os.path.join(temp_dir.name, FILES_DIR)
        os.makedirs(local_files_dir, exist_ok=True)

        # save plot locally
        plot_artifact = self._save_plot(key=key, step=step, local_dir=local_files_dir)

        # save internal metadata
        internal_metadata = PlotVersionInternalMetadata(
            files_dir=FILES_DIR,
            plot_file=posixpath.join(FILES_DIR, plot_artifact.artifact_file),
        )
        local_internal_metadata_path = os.path.join(
            temp_dir.name, INTERNAL_METADATA_PATH
        )
        os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
        with open(local_internal_metadata_path, "w") as f:
            json.dump(internal_metadata.dict(), f)

        # log the artifact
        _log_artifact_version_helper(
            run=run,
            name=key,
            artifact_type=ArtifactType.PLOT,
            artifact_dir=temp_dir,
            step=step,
        )


class PlotRunLogType(PydanticBase):
    value: PlotArtifact

    @staticmethod
    def get_log_type() -> str:
        return "plot"


class PlotVersionInternalMetadata(ArtifactVersionInternalMetadata):
    plot_file: str


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    import plotly.express as px

    import mlfoundry

    client = mlfoundry.get_client()

    run = client.create_run(ml_repo="plot-test1")

    df = px.data.iris()
    fig = px.scatter(
        df,
        x="sepal_width",
        y="sepal_length",
        color="species",
        size="petal_length",
        hover_data=["petal_width"],
    )
    Plot(fig).save(run, "foo")

    df = px.data.tips()
    fig = px.histogram(
        df,
        x="total_bill",
        y="tip",
        color="sex",
        marginal="rug",
        hover_data=df.columns,
    )
    Plot(fig).save(run, "foo", step=1)

    names = ["group_a", "group_b", "group_c"]
    values = [1, 10, 100]
    plt.figure(figsize=(9, 3))
    plt.subplot(131)
    plt.bar(names, values)
    plt.subplot(132)
    plt.scatter(names, values)
    plt.subplot(133)
    plt.plot(names, values)
    plt.suptitle("Categorical Plotting")

    Plot(plt).save(run, "bar")

    plt.clf()

    data = {
        "a": np.arange(50),
        "c": np.random.randint(0, 50, 50),
        "d": np.random.randn(50),
    }
    data["b"] = data["a"] + 10 * np.random.randn(50)
    data["d"] = np.abs(data["d"]) * 100
    plt.scatter("a", "b", c="c", s="d", data=data)
    plt.xlabel("entry a")
    plt.ylabel("entry b")
    Plot(plt).save(run, "bar", 2)

    plt.clf()

    ax = plt.subplot()
    t = np.arange(0.0, 5.0, 0.01)
    s = np.cos(2 * np.pi * t)
    (line,) = plt.plot(t, s, lw=2)

    plt.annotate(
        "local max",
        xy=(2, 1),
        xytext=(3, 1.5),
        arrowprops=dict(facecolor="black", shrink=0.05),
    )

    plt.ylim(-2, 2)
    Plot(plt.gcf()).save(run, "bar", 3)



import json
import os
import posixpath
import shutil
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Union

from mlflow.entities import ArtifactType

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.artifact import (
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from mlfoundry.log_types.artifacts.constants import FILES_DIR, INTERNAL_METADATA_PATH
from mlfoundry.log_types.image.constants import (
    DEFAULT_IMAGE_FORMAT,
    IMAGE_KEY_REGEX,
    IMAGE_METADATA_FILE_NAME,
    MISSING_PILLOW_PACKAGE_MESSAGE,
)
from mlfoundry.log_types.image.image_normalizer import normalize_image
from mlfoundry.log_types.image.types import BoundingBoxGroups, ClassGroups
from mlfoundry.log_types.pydantic_base import PydanticBase
from mlfoundry.run_utils import get_module

DataOrPathType = Union[str, Path, "numpy.ndarray", "PIL.Image.Image"]
ClassGroupsType = Dict[str, Union[str, List[str]]]
BBoxGrouptype = Dict[str, List[Dict]]


def validate_key_name(key: str):
    if not key or not IMAGE_KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run image key: {key} should only contain alphanumeric, hyphen or underscore"
        )


class ImageVersionInternalMetadata(ArtifactVersionInternalMetadata):
    image_file: str
    image_metadata_file: str


class Image:
    def __init__(
        self,
        data_or_path: DataOrPathType,
        caption: Optional[str] = None,
        class_groups: Optional[ClassGroupsType] = None,
        bbox_groups: Optional[BBoxGrouptype] = None,
    ):
        """Represent and log image using this class in `mlfoundry`.

        You can initialize `mlfoundry.Image` by either by using a local path
        or you can use a numpy array / PIL.Image object.

        If you are using numpy array, we only support the following data types,
        - bool
        - integer [0 - 255]
        - unsigned integer [0 - 255]
        - float [0.0 - 1.0]

        Any out of range value will be clipped.

        As for array shape/dim, we follow the following structures,
        - H x W (Grayscale)
        - H x W x 1 (Grayscale)
        - H x W x 3 (an RGB channel order is assumed)
        - H x W x 4 (an RGBA channel order is assumed)

        `PIL` package is required to log images. To install the `PIL` package,
        run `pip install pillow`.

        We can also log class names and bounding boxes associated with the image.
        Class names and bounding boxes should be always grouped under `actuals` or
        `predictions`. For example, if we have an image where the ground truth class
        is "cat" and predicted class is "dog", we can represent it like,

        ```python
        mlfoundry.Image(
            data_or_path=imarray,
            class_groups={"actuals": "dog", "predictions": "cat"}
        )
        ```

        You can define a bounding box using the following dictionary structure,
        ```python
        {
            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30}, # required, defines the position of the bounding box
                                                                             # (min_x, min_y) defines the top left and
                                                                             # (max_x, max_y) defines the bottom right corner of the box.
            "class_name": "dog", # required, the class name of the bounding box
            "caption": "dog", # optional, the caption of the bounding box.
                              # If not passed, the class name is set as caption.
        }
        ```

        Args:
            data_or_path (Union[str, Path, "numpy.ndarray", "PIL.Image.Image"]):
                Either the local path or the image object (Numpy array or PIL Image).
            caption (Optional[str], optional): A string caption or label for the image.
            class_groups (Optional[Dict[str, Union[str, List[str]]]], optional):
                Class names associated with the image. Expects class name(s) grouped by
                `predictions` or `actuals`.
            bbox_groups (Optional[Dict[str, List[Dict]]], optional): Bounding boxes
                associated with the image. Expects bounding boxes grouped by `predictions`
                or `actuals`.

        Examples:
        ### Logging images with caption and class names

        ```python
        import mlfoundry
        import numpy as np

        client = mlfoundry.get_client()
        run = client.create_run(
            ml_repo="my-classification-project",
        )

        imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))

        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": "dog", "predictions": "cat"},
            ),
        }
        run.log_images(images_to_log, step=1)

        run.end()
        ```

        ### Logging images for a multi-label classification problem

        ```python
        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": ["dog", "human"], "predictions": ["cat", "human"]},
            ),
        }

        run.log_images(images_to_log, step=1)
        ```

        ### Logging images with bounding boxes

        ```python
        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                bbox_groups={
                    "predictions": [
                        {
                            "position": {"min_x": 5, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "cat",
                        }
                    ],
                    "actuals": [
                        {
                            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "dog",
                            "caption": "dog",
                        }
                    ],
                },
            ),
        }

        run.log_images(images_to_log, step=1)
        ```
        """

        self._caption = None
        self._class_groups = None
        self._bbox_groups = None

        self._image = None
        self._image_artifact_path = None
        self._local_image_path = None

        self._init_image(data_or_path)
        self._init_metadata(
            caption=caption, class_groups=class_groups, bbox_groups=bbox_groups
        )

    @property
    def image(self) -> "PIL.Image.Image":
        if self._image is None:
            raise MlFoundryException("Image is not initialized")
        return self._image

    @property
    def caption(self) -> Optional[str]:
        return self._caption

    @property
    def class_groups(self) -> Optional[ClassGroups]:
        return self._class_groups

    @property
    def bbox_groups(self) -> Optional[BoundingBoxGroups]:
        return self._bbox_groups

    def save_image_to_local_dir(self, local_dir) -> str:
        if self._local_image_path is None:
            image_file_name = self._serialize_and_save_image_as_artifact(local_dir)
        else:
            image_file_name = self._copy_local_image_as_artifact(local_dir)

        return image_file_name

    def _serialize_and_save_image_as_artifact(self, local_dir: str) -> str:
        file_name = f"image.{DEFAULT_IMAGE_FORMAT}"
        local_path = os.path.join(local_dir, file_name)
        self.image.save(local_path)

        return file_name

    def _copy_local_image_as_artifact(self, local_dir: str) -> str:
        file_name = os.path.basename(self._local_image_path)
        new_local_image_path = os.path.join(local_dir, file_name)
        shutil.copy2(self._local_image_path, new_local_image_path)
        return file_name

    def _init_image(self, data_or_path: DataOrPathType):
        pil_image_module = get_module(
            module_name="PIL.Image",
            required=True,
            error_message=MISSING_PILLOW_PACKAGE_MESSAGE,
        )
        if isinstance(data_or_path, (str, Path)):
            self._local_image_path = os.path.abspath(data_or_path)
            with pil_image_module.open(data_or_path) as image:
                image.load()
                self._image = image
        else:
            self._image = normalize_image(data_or_path)

    def _init_metadata(
        self,
        caption: Optional[str],
        class_groups: Optional[ClassGroupsType],
        bbox_groups: Optional[Dict],
    ):
        if caption is not None:
            self._caption = str(caption)

        if class_groups:
            self._class_groups = ClassGroups.parse_obj(class_groups)

        if bbox_groups:
            self._bbox_groups = BoundingBoxGroups.parse_obj(bbox_groups)

    def _to_dict(self) -> Dict:
        dict_ = {}
        dict_["caption"] = self._caption
        dict_["class_groups"] = (
            self.class_groups.to_dict() if self.class_groups is not None else None
        )
        dict_["bbox_groups"] = (
            self.bbox_groups.to_dict() if self.bbox_groups is not None else None
        )

        return dict_

    def save(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
    ):
        validate_key_name(key)

        # creating temp dir, files directory, .truefoundry directory
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

        local_internal_metadata_path = os.path.join(
            temp_dir.name, INTERNAL_METADATA_PATH
        )
        os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)

        local_files_dir = os.path.join(temp_dir.name, FILES_DIR)
        os.makedirs(local_files_dir, exist_ok=True)

        try:
            # saving the image file
            image_file_name = self.save_image_to_local_dir(local_files_dir)

            # saving the image_metadata
            local_image_metadata_file_path = os.path.join(
                local_files_dir, IMAGE_METADATA_FILE_NAME
            )
            with open(local_image_metadata_file_path, "w") as fp:
                fp.write(ImageRunLogType(value=self._to_dict()).json())

            # creating and saving the internal_metadata file
            internal_metadata = ImageVersionInternalMetadata(
                files_dir=FILES_DIR,
                # joining with posixpath for avoiding backslash in case logged from windows machine
                image_file=posixpath.join(FILES_DIR, image_file_name),
                image_metadata_file=posixpath.join(FILES_DIR, IMAGE_METADATA_FILE_NAME),
            )
            with open(local_internal_metadata_path, "w") as f:
                json.dump(internal_metadata.dict(), f)

        except Exception as e:
            temp_dir.cleanup()
            raise MlFoundryException("Failed to log Image") from e

        _log_artifact_version_helper(
            run=run,
            name=key,
            artifact_type=ArtifactType.IMAGE,
            artifact_dir=temp_dir,
            step=step,
        )


class ImageRunLogType(PydanticBase):
    value: Dict

    @staticmethod
    def get_log_type() -> str:
        return "image"



import copy
import datetime
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

from mlflow.entities import Artifact, ArtifactType
from mlflow.entities import ArtifactVersion as _ArtifactVersion
from mlflow.tracking import MlflowClient

from mlfoundry.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import INTERNAL_METADATA_PATH
from mlfoundry.log_types.artifacts.utils import (
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
    calculate_local_directory_size,
)
from mlfoundry.logger import logger
from mlfoundry.pydantic_v1 import BaseModel, Extra

# TODO: Add some progress indicators for upload and download
# TODO: Support async download and upload


class ArtifactPath(NamedTuple):
    src: str
    dest: str = None


class ArtifactVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root


class ArtifactVersionDownloadInfo(BaseModel):
    download_dir: str
    content_dir: str


class ArtifactVersion:
    def __init__(self, artifact_version: _ArtifactVersion, artifact: Artifact) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._artifact_version: _ArtifactVersion = artifact_version
        self._artifact: Artifact = artifact
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str) -> "ArtifactVersion":
        """
        Get the version of an Artifact to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Returns:
            ArtifactVersion: An ArtifactVersion instance of the artifact

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = mlfoundry.ArtifactVersion.from_fqn(fqn="<artifact-fqn>")
            ```
        """
        mlflow_client = _get_mlflow_client()
        artifact_version = mlflow_client.get_artifact_version_by_fqn(fqn=fqn)
        artifact = mlflow_client.get_artifact_by_id(
            artifact_id=artifact_version.artifact_id
        )
        return cls(artifact_version=artifact_version, artifact=artifact)

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Artifact Version was deleted, cannot access a deleted version"
            )

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._artifact_version = self._mlflow_client.get_artifact_version_by_id(
                version_id=self._artifact_version.id
            )
        self._description = self._artifact_version.description or ""
        self._metadata = copy.deepcopy(self._artifact_version.artifact_metadata)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=self._artifact_version.id
            ),
            mlflow_client=self._mlflow_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the artifact"""
        return self._artifact.name

    @property
    def artifact_fqn(self) -> str:
        """Get fqn of the artifact"""
        return self._artifact.fqn

    @property
    def version(self) -> int:
        """Get version information of the artifact"""
        return self._artifact_version.version

    @property
    def fqn(self) -> str:
        """Get fqn of the current artifact version"""
        return self._artifact_version.fqn

    @property
    def step(self) -> int:
        """Get the step in which artifact was created"""
        return self._artifact_version.step

    @property
    def description(self) -> Optional[str]:
        """Get description of the artifact"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the artifact"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current artifact"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current artifact"""
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def created_by(self) -> str:
        """Get the information about who created the artifact"""
        return self._artifact_version.created_by

    @property
    def created_at(self) -> datetime.datetime:
        """Get the time at which artifact was created"""
        return self._artifact_version.created_at

    @property
    def updated_at(self) -> datetime.datetime:
        """Get the information about the when the artifact was updated"""
        return self._artifact_version.updated_at

    def raw_download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified artifacts. This directory must already exist.
                        If unspecified, the artifacts will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired artifacts.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.raw_download(path="<your-desired-download-path>")
            ```
        """
        logger.info(
            "Downloading artifact version contents, this might take a while ..."
        )
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.download_artifacts(
            artifact_path="", dst_path=path, overwrite=overwrite, progress=progress
        )

    def _download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> Tuple[ArtifactVersionInternalMetadata, str]:
        self._ensure_not_deleted()
        download_dir = self.raw_download(
            path=path, overwrite=overwrite, progress=progress
        )
        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                f"Artifact version seems to be corrupted or in invalid format due to missing artifact metadata. "
                f"You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ArtifactVersionInternalMetadata.parse_obj(json.load(f))
        download_path = os.path.join(download_dir, internal_metadata.files_dir)
        return internal_metadata, download_path

    def download(
        self,
        path: Optional[Union[str, Path]] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified artifacts. This directory must already exist.
                        If unspecified, the artifacts will either be downloaded to a new
                        uniquely-named directory on the local filesystem or will be returned
                        directly in the case of the Local ArtifactRepository.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired artifacts.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.download(path="<your-desired-download-path>")
            ```
        """
        _, download_path = self._download(
            path=path, overwrite=overwrite, progress=progress
        )
        return download_path

    def delete(self) -> bool:
        """
        Deletes the current instance of the ArtifactVersion hence deleting the current version.

        Returns:
            True if artifact was deleted successfully

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.delete()
            ```
        """
        self._ensure_not_deleted()
        self._mlflow_client.delete_artifact_version(
            version_id=self._artifact_version.id
        )
        self._deleted = True
        return True

    def update(self):
        """
        Updates the current instance of the ArtifactVersion hence updating the current version.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.description = 'This is the new description'
            artifact_version.update()
            ```
        """
        self._ensure_not_deleted()

        self._artifact_version = self._mlflow_client.update_artifact_version(
            version_id=self._artifact_version.id,
            description=self.description,
            artifact_metadata=self.metadata,
        )
        self._set_mutable_attrs()


def _log_artifact_version_helper(
    run,
    name: str,
    artifact_type: ArtifactType,
    artifact_dir: tempfile.TemporaryDirectory,
    mlflow_client: Optional[MlflowClient] = None,
    ml_repo_id: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
    progress: Optional[bool] = None,
):
    if (run and mlflow_client) or (not run and not mlflow_client):
        raise MlFoundryException("Exactly one of run, mlflow_client should be passed")
    if mlflow_client and not ml_repo_id:
        raise MlFoundryException(
            "If mlflow_client is passed, ml_repo_id must also be passed"
        )
    if run:
        mlflow_client: MlflowClient = run.mlflow_client

    version_id = mlflow_client.create_artifact_version(
        experiment_id=int(run._experiment_id) if run else ml_repo_id,
        artifact_type=artifact_type,
        name=name,
    )
    artifacts_repo = MlFoundryArtifactsRepository(
        artifact_identifier=ArtifactIdentifier(artifact_version_id=version_id),
        mlflow_client=mlflow_client,
    )
    total_size = calculate_local_directory_size(artifact_dir)
    try:
        logger.info(
            "Packaging and uploading files to remote with Artifact Size: %.6f MB",
            total_size / 1000000.0,
        )
        artifacts_repo.log_artifacts(
            local_dir=artifact_dir.name, artifact_path=None, progress=progress
        )
    except Exception as e:
        mlflow_client.notify_failure_for_artifact_version(version_id=version_id)
        raise MlFoundryException("Failed to log Artifact") from e
    finally:
        artifact_dir.cleanup()
    artifact_version = mlflow_client.finalize_artifact_version(
        version_id=version_id,
        run_uuid=run.run_id if run else None,
        description=description,
        artifact_metadata=metadata,
        data_path=INTERNAL_METADATA_PATH,
        step=step,
        artifact_size=total_size,
    )

    return ArtifactVersion.from_fqn(fqn=artifact_version.fqn)



import queue
import threading
import time
import typing

from mlfoundry.logger import logger
from mlfoundry.monitoring.entities import BasePacket
from mlfoundry.monitoring.store.repositories import RestMonitoringStore


class MonitoringStoreWorker(threading.Thread):
    def __init__(
        self,
        task_queue: queue.Queue,
        terminate_event: threading.Event,
        monitoring_store: RestMonitoringStore,
        flush_interval: float,
        flush_every_num_message: int,
    ):
        super().__init__(daemon=True)
        self.task_queue: "queue.Queue[BasePacket]" = task_queue
        self.terminate_event = terminate_event
        self.monitoring_store: RestMonitoringStore = monitoring_store
        self.flush_interval: float = flush_interval
        self.flush_every_num_message: int = flush_every_num_message

    def next_batch_from_queue(self) -> typing.List[BasePacket]:
        batch = []

        start_time = time.monotonic()
        time_to_stop = start_time + self.flush_interval

        while len(batch) < self.flush_every_num_message:
            time_delta = time_to_stop - time.monotonic()
            if time_delta <= 0:
                logger.debug("flushing as time interval has reached")
                break
            try:
                inference_packet = self.task_queue.get(block=True, timeout=time_delta)
                batch.append(inference_packet)
            except queue.Empty:
                logger.debug("task queue is empty")

        return batch

    def work(self):
        inference_packets = self.next_batch_from_queue()
        if len(inference_packets) == 0:
            return
        try:
            self.monitoring_store.batch_log_inference(inference_packets)
        except Exception:
            logger.exception(
                "fail to log monitoring\n"
                f"discarding {len(inference_packets)} inference_packets"
            )
        for _ in range(len(inference_packets)):
            self.task_queue.task_done()

    def run(self):
        logger.debug("worker started")
        while not self.terminate_event.is_set():
            self.work()
        logger.debug("worker exited")



import typing


def get_class_names(predictions, actuals) -> typing.List:
    class_names = set()
    if predictions is not None:
        class_names.update(predictions)
    if actuals is not None:
        class_names.update(actuals)
    class_names = sorted(class_names)
    return class_names



import collections
import copy
import datetime
import tempfile
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from mlflow.entities import Dataset as _Dataset

from mlfoundry.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.utils import (
    _copy_additional_files,
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
    calculate_local_directory_size,
)
from mlfoundry.logger import logger


class DataDirectoryPath(NamedTuple):
    src: str
    dest: str = None


class DataDirectory:
    def __init__(self, dataset: _Dataset) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._dataset: _Dataset = dataset
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str):
        """
        Get the DataDirectory to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the data directory.

        Returns:
            DataDirectory: An DataDirectory instance of the artifact

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            data_directory = mlfoundry.DataDirectory.from_fqn(fqn="<data_directory-fqn>")
            ```
        """
        mlflow_client = _get_mlflow_client()
        dataset = mlflow_client.get_dataset_by_fqn(fqn=fqn)
        return cls(dataset)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._dataset = self._mlflow_client.get_dataset_by_fqn(
                fqn=self._dataset.fqn
            )
        self._description = self._dataset.description or ""
        self._metadata = copy.deepcopy(self._dataset.dataset_metadata)

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(dataset_fqn=self._dataset.fqn),
            mlflow_client=self._mlflow_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the DataDirectory"""
        return self._dataset.name

    @property
    def fqn(self) -> str:
        """Get fqn of the DataDirectory"""
        return self._dataset.fqn

    @property
    def storage_root(self) -> str:
        """Get storage_root of the DataDirectory"""
        return self._dataset.storage_root

    @property
    def description(self) -> Optional[str]:
        """Get description of the DataDirectory"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the DataDirectory"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current DataDirectory"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current DataDirectory"""
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def created_by(self) -> str:
        """Get the information about who created the DataDirectory"""
        return self._dataset.created_by

    @property
    def created_at(self) -> datetime.datetime:
        """Get the time at which DataDirectory was created"""
        return self._dataset.created_at

    @property
    def updated_at(self) -> datetime.datetime:
        """Get the information about the when the DataDirectory was updated"""
        return self._dataset.updated_at

    def add_files(
        self,
        file_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], DataDirectoryPath]
        ],
    ) -> None:
        """Logs File in the `DataDirectory`.

        Args:
            file_paths (List[mlfoundry.DataDirectoryPath], optional): A list of pairs of (source path, destination path) to add files and folders to the DataDirectory contents. The first member of the pair should be a file or directory path and the second member should be the path inside the artifact contents to upload to.

                ```
                E.g. >>> data_directory.add_files(
                     ...     file_paths=[
                                mlfoundry.DataDirectoryPath("foo.txt", "foo/bar/foo.txt"),
                                mlfoundry.DataDirectoryPath("tokenizer/", "foo/tokenizer/"),
                                mlfoundry.DataDirectoryPath('bar.text'),
                                ('bar.txt', ),
                                ('foo.txt', 'a/foo.txt')
                             ]
                     ... )
                would result in
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```

        Examples:
            ```python
            import os
            import mlfoundry

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = mlfoundry.get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<data_directory-fqn>")

            data_directory.add_files(
                file_paths=[mlfoundry.DataDirectoryPath('artifact.txt', 'a/b/')]
            )
            ```
        """
        for i, file_path in enumerate(file_paths):
            if isinstance(file_path, DataDirectoryPath):
                continue
            elif isinstance(file_path, collections.abc.Sequence) and (
                0 < len(file_path) <= 2
            ):
                file_paths[i] = DataDirectoryPath(*file_path)
            else:
                raise ValueError(
                    "`file_paths` should be an instance of `mlfoundry.DataDirectoryPath` or a tuple of (src, dest) path strings"
                )

        logger.info("Adding the files to data_directory, this might take a while ...")
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

        try:
            logger.info("Copying the files to add")
            _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir="",
                model_dir=None,
                additional_files=file_paths,
            )

        except Exception as e:
            temp_dir.cleanup()
            raise MlFoundryException("Failed to Add Files to DataDirectory") from e

        artifacts_repo = self._get_artifacts_repo()
        total_size = calculate_local_directory_size(temp_dir)
        try:
            logger.info(
                "Packaging and uploading files to remote with Size: %.6f MB",
                total_size / 1000000.0,
            )
            artifacts_repo.log_artifacts(local_dir=temp_dir.name, artifact_path=None)
        except Exception as e:
            raise MlFoundryException("Failed to Add Files to DataDirectory") from e
        finally:
            temp_dir.cleanup()

    def list_files(
        self,
        path: Optional[str] = None,
    ):
        """
        List all the files and folders in the data_directory.

        Args:
            path: The path of directory in data_directory, from which the files are to be listed.

        Returns:
            str: Absolute path of the local filesystem location containing the desired files or folder.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-artifact-fqn>")
            files = data_directory.list_files()
            for file in files:
                print(file.path)
            ```
        """
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.list_artifacts(path=path)

    def download(
        self,
        download_path: Optional[str] = "",
        path: Optional[str] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ):
        """
        Download an file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            download_path: Relative source path to the desired files.
            path: Absolute path of the local filesystem destination directory to which to
                download the specified files. This directory must already exist.
                If unspecified, the files will either be downloaded to a new
                uniquely-named directory.
            overwrite: if to overwrite the files at/inside `dst_path` if they exist
            progress: value to show progress bar, defaults to None.

        Returns:
            str: Absolute path of the local filesystem location containing the desired files or folder.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-data_directory-fqn>")
            data_directory.download(download_path="<your-desired-download-path>")
            ```
        """
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.download_artifacts(
            artifact_path=download_path,
            dst_path=path,
            overwrite=overwrite,
            progress=progress,
        )

    def update(self):
        """
        Updates the current instance of the DataDirectory.

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-artifact-fqn>")
            data_directory.description = 'This is the new description'
            data_directory.update()
            ```
        """

        self._dataset = self._mlflow_client.update_dataset(
            fqn=self._dataset.fqn,
            description=self.description,
            dataset_metadata=self.metadata,
        )
        self._set_mutable_attrs()



from dataclasses import dataclass
from typing import List

from mlflow.entities import Metric


@dataclass
class Event(object):
    """
    Base Event type
    """

    producer: str


@dataclass
class LogMetricsEvent(Event):
    """
    Metrics Event
    """

    metrics: List[Metric]



from mlflow.entities import Feature

from mlfoundry.enums import *
from mlfoundry.log_types import Image, Plot
from mlfoundry.log_types.artifacts.artifact import ArtifactPath, ArtifactVersion
from mlfoundry.log_types.artifacts.dataset import DataDirectory, DataDirectoryPath
from mlfoundry.log_types.artifacts.model import CustomMetric, ModelSchema, ModelVersion
from mlfoundry.logger import init_logger
from mlfoundry.login import login
from mlfoundry.mlfoundry_api import get_client
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.monitoring.entities import Actual, Prediction, PredictionData
from mlfoundry.version import __version__

__all__ = [
    "FileFormat",
    "ModelFramework",
    "DataSlice",
    "ModelType",
    "get_client",
    "__version__",
    "MlFoundryRun",
    "login",
    "Image",
    "Plot",
    "Prediction",
    "PredictionData",
    "Actual",
    "Feature",
    "ModelSchema",
    "ModelVersion",
    "CustomMetric",
    "ArtifactPath",
    "ArtifactVersion",
    "DataDirectory",
    "DataDirectoryPath",
]


init_logger()
